import java.util.Iterator;
import java.util.NoSuchElementException;

// Note. On pourrait tenter d'implémenter un itérateur sur un Tree
// en créant un itérateur sur le fils gauche et un itérateur sur le
// fils droit, et en présentant au client les éléments du premier
// itérateur, puis la valeur [val], puis les éléments du second
// itérateur. Cela serait en fait tout-à-fait analogue à la fonction
// elements : 'a tree -> 'a list qui produit la liste des éléments
// d'un arbre en concaténant la liste des éléments du fils gauche,
// l'élément [val], et la liste des éléments du fils droit. La
// complexité de cette approche est mauvaise: O(nlog n) si l'arbre
// est équilibré, O(n^2) s'il est déséquilibré vers la gauche.

// On pourrait modifier cette approche pour qu'elle fonctionne en
// temps O(n). Il faudrait pour cela introduire un paramètre accu
// (une liste d'arbres). Cela ressemblerait à ce qui a été fait au
// TD06 (arbres en OCaml). Cette liste occuperait un espace
// auxiliaire O(log n).

// Les solutions évoquées ci-dessus n'exploitent pas le champ parent,
// et sont sous-optimales. En exploitant le champ parent, il existe une
// solution en temps O(n) et en espace auxiliaire O(1). La voici.

class InorderIterator<V> implements Iterator<V> {

  private Tree<V> current;
  // contient le prochain noeud du parcours
  // en particulier, le sous-arbre gauche a déjà été parcouru

  public InorderIterator(Tree<V> t) {
    current = leftmost(t);
  }

  public boolean hasNext() {
    return current != null;
  }

  public V next() {
    if (!hasNext()) throw new NoSuchElementException();
    // la valeur à renvoyer est donnée par current
    V v = current.val;
    // il faut maintenant mettre current à jour
    if (current.right != null) {
      // s'il existe un sous-arbre droit, on y passe
      current = leftmost(current.right);
    } else {
      // sinon, on remonte tant qu'on est un sous-arbre droit
      while (current.parent != null && current == current.parent.right) {
        current = current.parent;
      }
      // puis on remonte d'un cran
      current = current.parent;
    }
    return v;
  }

  // descend tout en bas à gauche
  static<V> Tree<V> leftmost(Tree<V> t) {
    while (t.left != null)
      t = t.left;
    return t;
  }

}
